from importlib import import_module

from django.apps import apps
from django.conf import settings
from django.urls import include, path, re_path
from django.views.decorators.cache import never_cache
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import TemplateView
from django.views.i18n import JavaScriptCatalog

import calendarweek.django
from ckeditor_uploader import views as ckeditor_uploader_views
from health_check.urls import urlpatterns as health_urls
from maintenance_mode.decorators import force_maintenance_mode_off
from oauth2_provider.views import ConnectDiscoveryInfoView
from rules.contrib.views import permission_required

from . import views

urlpatterns = [
    path(
        "",
        force_maintenance_mode_off(TemplateView.as_view(template_name="core/vue_index.html")),
        name="vue_app",
    ),
    path("manifest.json", views.ManifestView.as_view(), name="manifest"),
    path("sw.js", views.ServiceWorkerView.as_view(), name="service_worker"),
    path(settings.MEDIA_URL.removeprefix("/"), include("titofisto.urls")),
    path("__icons__/", include("dj_iconify.urls")),
    path(
        "graphql/",
        csrf_exempt(views.LoggingGraphQLView.as_view()),
        name="graphql",
    ),
    path("logo", force_maintenance_mode_off(views.LogoView.as_view()), name="logo"),
    path(
        ".well-known/openid-configuration",
        ConnectDiscoveryInfoView.as_view(),
        name="oidc_configuration",
    ),
    path(
        ".well-known/carddav/",
        views.DAVWellKnownView.as_view(),
        name="wellknown_carddav",
    ),
    path(
        ".well-known/caldav/",
        views.DAVWellKnownView.as_view(),
        name="wellknown_caldav",
    ),
    path("oauth/applications/", views.TemplateView.as_view(template_name="core/vue_index.html")),
    path(
        "oauth/applications/register/",
        views.TemplateView.as_view(template_name="core/vue_index.html"),
    ),
    path(
        "oauth/applications/<int:pk>/",
        views.TemplateView.as_view(template_name="core/vue_index.html"),
    ),
    path(
        "oauth/applications/<int:pk>/delete/",
        views.TemplateView.as_view(template_name="core/vue_index.html"),
    ),
    path(
        "oauth/applications/<int:pk>/edit/",
        views.TemplateView.as_view(template_name="core/vue_index.html"),
    ),
    path(
        "oauth/authorized_tokens/", views.TemplateView.as_view(template_name="core/vue_index.html")
    ),
    path("oauth/", include("oauth2_provider.urls", namespace="oauth2_provider")),
    path("system_status/", views.SystemStatusAPIView.as_view(), name="system_status_api"),
    path(
        "o/<str:app_label>/<str:model>/<uuid:pk>",
        views.ObjectRepresentationView.as_view(),
        name="object_representation_with_uuid",
    ),
    path(
        "o/<str:app_label>/<str:model>/<int:pk>",
        views.ObjectRepresentationView.as_view(),
        name="object_representation_with_pk",
    ),
    path(
        "o/",
        views.ObjectRepresentationView.as_view(),
        name="object_representation_anonymous",
    ),
    path(
        "feeds/<str:subregistry>/<str:name>.ics", views.ICalFeedView.as_view(), name="calendar_feed"
    ),
    path("feeds/<str:name>.ics", views.ICalAllFeedsView.as_view(), name="all_calendar_feeds"),
    path("dav/", views.DAVResourceView.as_view(), name="dav_registry"),
    path("dav/<str:name>/", views.DAVResourceView.as_view(), name="dav_subregistry"),
    path(
        "dav/<str:subregistry>/<str:name>/",
        views.DAVResourceView.as_view(),
        name="dav_resource",
    ),
    path(
        "dav/<str:subregistry>/<str:name>/<int:id>.ics",
        views.DAVSingleResourceView.as_view(),
        name="dav_resource_calendar",
    ),
    path(
        "dav/<str:subregistry>/<str:name>/<int:id>.vcf",
        views.DAVSingleResourceView.as_view(),
        name="dav_resource_contact",
    ),
    path("", include("django_prometheus.urls")),
    path(
        "django/",
        include(
            [
                path("accounts/login/", never_cache(views.LoginView.as_view()), name="login"),
                path("accounts/logout/", views.CustomLogoutView.as_view(), name="logout"),
                path(
                    "accounts/password/change/",
                    views.TemplateView.as_view(template_name="core/vue_index.html"),
                    name="account_change_password",
                ),
                path(
                    "accounts/password/reset/",
                    views.TemplateView.as_view(template_name="core/vue_index.html"),
                    name="account_reset_password",
                ),
                path(
                    "password/reset/done/",
                    views.TemplateView.as_view(template_name="core/vue_index.html"),
                    name="account_reset_password_done",
                ),
                path("accounts/", include("allauth.urls")),
                path(
                    "accounts/3rdparty/<int:pk>/delete/",
                    views.SocialAccountDeleteView.as_view(),
                    name="delete_social_account_by_pk",
                ),
                path("offline/", views.OfflineView.as_view(), name="offline"),
                path(
                    "invitations/send-invite/", views.InvitePerson.as_view(), name="invite_person"
                ),
                path(
                    "invitations/code/generate/",
                    views.GenerateInvitationCode.as_view(),
                    name="generate_invitation_code",
                ),
                path(
                    "invitations/disabled/",
                    views.InviteDisabledView.as_view(),
                    name="invite_disabled",
                ),
                path("invitations/", include("invitations.urls")),
                path(
                    "o/core/person/",
                    TemplateView.as_view(template_name="core/empty.html"),
                    name="persons",
                ),
                path(
                    "o/",
                    TemplateView.as_view(template_name="core/empty.html"),
                    name="person",
                ),
                path(
                    "o/core/person/<int:id_>",
                    TemplateView.as_view(template_name="core/empty.html"),
                    name="person_by_id",
                ),
                path(
                    "o/core/person/<int:pk>/invite/",
                    views.InvitePersonByID.as_view(),
                    name="invite_person_by_id",
                ),
                path(
                    "o/core/group/",
                    TemplateView.as_view(template_name="core/empty.html"),
                    name="groups",
                ),
                path("o/core/group/create/", views.edit_group, name="create_group"),
                path(
                    "o/core/group/<int:id_>",
                    TemplateView.as_view(template_name="core/empty.html"),
                    name="group_by_id",
                ),
                path("o/core/group/<int:id_>/edit/", views.edit_group, name="edit_group_by_id"),
                path("", TemplateView.as_view(template_name="core/vue_index.html"), name="index"),
                path("search/searchbar/", views.searchbar_snippets, name="searchbar_snippets"),
                path("search/", views.PermissionSearchView.as_view(), name="haystack_search"),
                path("maintenance-mode/", include("maintenance_mode.urls")),
                path("impersonate/", include("impersonate.urls")),
                path(
                    "oauth/authorize/",
                    views.CustomAuthorizationView.as_view(),
                    name="oauth2_provider:authorize",
                ),
                path("__i18n__/", include("django.conf.urls.i18n")),
                path(
                    "ckeditor/upload/",
                    permission_required("core.ckeditor_upload_files_rule")(
                        ckeditor_uploader_views.upload
                    ),
                    name="ckeditor_upload",
                ),
                path(
                    "ckeditor/browse/",
                    permission_required("core.ckeditor_upload_files_rule")(
                        ckeditor_uploader_views.browse
                    ),
                    name="ckeditor_browse",
                ),
                path("select2/", include("django_select2.urls")),
                path(
                    "calendarweek_i18n.js", calendarweek.django.i18n_js, name="calendarweek_i18n_js"
                ),
                path("gettext.js", JavaScriptCatalog.as_view(), name="javascript-catalog"),
                path(
                    "preferences/site/",
                    views.preferences,
                    {"registry_name": "site"},
                    name="preferences_site",
                ),
                path(
                    "preferences/person/",
                    views.preferences,
                    {"registry_name": "person"},
                    name="preferences_person",
                ),
                path(
                    "preferences/group/",
                    views.preferences,
                    {"registry_name": "group"},
                    name="preferences_group",
                ),
                path(
                    "preferences/person/<int:pk>/",
                    views.preferences,
                    {"registry_name": "person"},
                    name="preferences_person",
                ),
                path(
                    "preferences/group/<int:pk>/",
                    views.preferences,
                    {"registry_name": "group"},
                    name="preferences_group",
                ),
                path(
                    "preferences/person/<int:pk>/<str:section>/",
                    views.preferences,
                    {"registry_name": "person"},
                    name="preferences_person",
                ),
                path(
                    "preferences/group/<int:pk>/<str:section>/",
                    views.preferences,
                    {"registry_name": "group"},
                    name="preferences_group",
                ),
                path(
                    "preferences/site/<str:section>/",
                    views.preferences,
                    {"registry_name": "site"},
                    name="preferences_site",
                ),
                path(
                    "preferences/person/<str:section>/",
                    views.preferences,
                    {"registry_name": "person"},
                    name="preferences_person",
                ),
                path(
                    "preferences/group/<str:section>/",
                    views.preferences,
                    {"registry_name": "group"},
                    name="preferences_group",
                ),
                path("health/pdf/", views.TestPDFGenerationView.as_view(), name="test_pdf"),
                path("health/", include(health_urls)),
                path(
                    "permissions/global/user/",
                    views.UserGlobalPermissionsListBaseView.as_view(),
                    name="manage_user_global_permissions",
                ),
                path(
                    "permissions/global/group/",
                    views.GroupGlobalPermissionsListBaseView.as_view(),
                    name="manage_group_global_permissions",
                ),
                path(
                    "permissions/object/user/",
                    views.UserObjectPermissionsListBaseView.as_view(),
                    name="manage_user_object_permissions",
                ),
                path(
                    "permissions/object/group/",
                    views.GroupObjectPermissionsListBaseView.as_view(),
                    name="manage_group_object_permissions",
                ),
                path(
                    "permissions/global/user/<int:pk>/delete/",
                    views.UserGlobalPermissionDeleteView.as_view(),
                    name="delete_user_global_permission",
                ),
                path(
                    "permissions/global/group/<int:pk>/delete/",
                    views.GroupGlobalPermissionDeleteView.as_view(),
                    name="delete_group_global_permission",
                ),
                path(
                    "permissions/object/user/<int:pk>/delete/",
                    views.UserObjectPermissionDeleteView.as_view(),
                    name="delete_user_object_permission",
                ),
                path(
                    "permissions/object/group/<int:pk>/delete/",
                    views.GroupObjectPermissionDeleteView.as_view(),
                    name="delete_group_object_permission",
                ),
                path(
                    "permissions/assign/",
                    views.SelectPermissionForAssignView.as_view(),
                    name="select_permission_for_assign",
                ),
                path(
                    "permissions/<int:pk>/assign/",
                    views.AssignPermissionView.as_view(),
                    name="assign_permission",
                ),
            ]
        ),
    ),
]

# Use custom server error handler to get a request object in the template
handler500 = views.server_error

# Automatically mount URLs from all installed AlekSIS apps
for app_config in apps.app_configs.values():
    if not app_config.name.startswith("aleksis.apps."):
        continue

    try:
        urls_module = import_module(f"{app_config.name}.urls")
    except ModuleNotFoundError:
        # Ignore exception as app just has no URLs
        urls_module = None

    if hasattr(urls_module, "urlpatterns"):
        urlpatterns.append(
            path(f"django/app/{app_config.label}/", include(urls_module.urlpatterns))
        )

    if hasattr(urls_module, "api_urlpatterns"):
        urlpatterns.append(path(f"app/{app_config.label}/", include(urls_module.api_urlpatterns)))

    if hasattr(urls_module, "wellknown_urlpatterns"):
        urlpatterns.append(path(".well-known/", include(urls_module.wellknown_urlpatterns)))

urlpatterns.append(
    re_path(
        r"^(?P<url>.*)/$",
        force_maintenance_mode_off(TemplateView.as_view(template_name="core/vue_index.html")),
        name="vue_app",
    )
)



from typing import Union

from django.contrib.auth import get_user_model
from django.contrib.auth.password_validation import validate_password
from django.core.exceptions import PermissionDenied, SuspiciousOperation, ValidationError
from django.db import IntegrityError, transaction
from django.db.models import Q
from django.utils import timezone
from django.utils.translation import gettext as _

import graphene
import graphene_django_optimizer
from graphene_django import DjangoObjectType
from graphene_file_upload.scalars import Upload
from invitations.views import accept_invitation

from ..filters import PersonFilter
from ..models import (
    Activity,
    Address,
    AddressType,
    DummyPerson,
    Person,
    PersonGroupThrough,
    PersonInvitation,
    PersonRelationship,
    Role,
)
from ..util.auth_helpers import custom_username_validators
from ..util.core_helpers import get_site_preferences, has_person
from .address import AddressType as GraphQLAddressType
from .base import (
    BaseBatchCreateMutation,
    BaseBatchDeleteMutation,
    BaseBatchPatchMutation,
    BaseObjectType,
    DjangoFilterMixin,
    FieldFileType,
    PermissionsTypeMixin,
)
from .group import PersonGroupThroughType
from .notification import NotificationType


class PersonPreferencesType(graphene.ObjectType):
    theme_design_mode = graphene.String()
    days_of_week = graphene.List(graphene.Int)

    def resolve_theme_design_mode(parent, info, **kwargs):
        return parent["theme__design"]

    @staticmethod
    def resolve_days_of_week(root, info, **kwargs):
        first_day = root["calendar__first_day_of_the_week"]

        if first_day == "default":
            first_day = get_site_preferences()["calendar__first_day_of_the_week"]

        first_day = int(first_day)

        days = list(map(str, range(7)))
        sorted_days = days[first_day:] + days[:first_day]

        return list(map(int, sorted_days))


class PersonRelationshipType(BaseObjectType):
    """GraphQL type for PersonRelationship"""

    class Meta:
        model = PersonRelationship
        fields = [
            "person",
            "of_person",
            "roles",
        ]


class PersonType(PermissionsTypeMixin, DjangoFilterMixin, DjangoObjectType):
    class Meta:
        model = Person
        fields = [
            "id",
            "user",
            "first_name",
            "last_name",
            "additional_name",
            "short_name",
            "addresses",
            "phone_number",
            "mobile_number",
            "email",
            "date_of_birth",
            "place_of_birth",
            "sex",
            "photo",
            "avatar",
            "primary_group",
            "description",
            "owner_of",
            "member_of",
            "related_persons",
            "related_to_persons",
        ]
        filterset_class = PersonFilter

    id = graphene.ID(required=False)

    full_name = graphene.String()
    username = graphene.String()
    userid = graphene.ID()
    photo = graphene.Field(FieldFileType, required=False)
    avatar = graphene.Field(FieldFileType, required=False)
    avatar_url = graphene.String()
    avatar_content_url = graphene.String()
    secondary_image_url = graphene.String(required=False)

    addresses = graphene.List(GraphQLAddressType, required=False)

    phone_number = graphene.String(required=False)
    mobile_number = graphene.String(required=False)
    email = graphene.String(required=False)

    date_of_birth = graphene.String(required=False)
    place_of_birth = graphene.String(required=False)

    notifications = graphene.List(NotificationType)
    unread_notifications_count = graphene.Int(required=False)

    is_dummy = graphene.Boolean()
    preferences = graphene.Field(PersonPreferencesType)

    can_change_person_preferences = graphene.Boolean()
    can_impersonate_person = graphene.Boolean()
    can_invite_person = graphene.Boolean()
    can_change_password = graphene.Boolean()
    can_send_password_reset_request = graphene.Boolean()

    person_relationships = graphene.List(PersonRelationshipType)
    group_relationships = graphene.List(PersonGroupThroughType)

    children = graphene.List(lambda: PersonType)
    guardians = graphene.List(lambda: PersonType)

    def resolve_addresses(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_addresses_rule", root):
            return root.addresses.all()
        return []

    def resolve_phone_number(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_contact_details_rule", root):
            return root.phone_number
        return None

    def resolve_mobile_number(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_contact_details_rule", root):
            return root.mobile_number
        return None

    def resolve_email(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_contact_details_rule", root):
            return root.email or (root.user.email if root.user else None)
        return None

    def resolve_date_of_birth(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_personal_details_rule", root):
            return root.date_of_birth
        return None

    def resolve_place_of_birth(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_personal_details_rule", root):
            return root.place_of_birth
        return None

    def resolve_person_relationships(root, info, **kwargs):  # noqa
        if getattr(root, "can_view_person_groups", False) or info.context.user.has_perm(
            "core.view_personal_details", root
        ):
            return PersonRelationship.objects.filter(Q(person=root) | Q(of_person=root))
        return []

    def resolve_related_persons(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_personal_details_rule", root):
            return root.related_persons.all()
        return []

    def resolve_related_to_persons(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_personal_details_rule", root):
            return root.related_to_persons.all()
        return []

    def resolve_children(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_personal_details_rule", root):
            relationships = PersonRelationship.objects.filter(
                person=root, roles__short_name="guardian"
            )
            return Person.objects.filter(pk__in=relationships.values_list("of_person", flat=True))
        return []

    def resolve_guardians(root, info, **kwargs):  # noqa
        if info.context.user.has_perm("core.view_personal_details_rule", root):
            relationships = PersonRelationship.objects.filter(
                of_person=root, roles__short_name="guardian"
            )
            return Person.objects.filter(pk__in=relationships.values_list("person", flat=True))
        return []

    def resolve_group_relationships(root, info, **kwargs):  # noqa
        if getattr(root, "can_view_person_groups", False) or info.context.user.has_perm(
            "core.view_person_groups_rule", root
        ):
            return PersonGroupThrough.objects.filter(person=root)
        return []

    def resolve_member_of(root, info, **kwargs):  # noqa
        if getattr(root, "can_view_person_groups", False) or info.context.user.has_perm(
            "core.view_person_groups_rule", root
        ):
            return root.member_of.all()
        return []

    def resolve_owner_of(root, info, **kwargs):  # noqa
        if getattr(root, "can_view_person_groups", False) or info.context.user.has_perm(
            "core.view_person_groups_rule", root
        ):
            return root.owner_of.all()
        return []

    @graphene_django_optimizer.resolver_hints(
        model_field="user",
    )
    def resolve_username(root, info, **kwargs):  # noqa
        return root.user.username if root.user else None

    @graphene_django_optimizer.resolver_hints(
        model_field="user",
    )
    def resolve_userid(root, info, **kwargs):  # noqa
        return root.user.id if root.user else None

    def resolve_unread_notifications_count(root, info, **kwargs):  # noqa
        if root.pk and has_person(info.context) and root == info.context.user.person:
            return root.unread_notifications_count
        elif root.pk:
            return 0
        return None

    def resolve_photo(root, info, **kwargs):
        if info.context.user.has_perm("core.view_photo_rule", root):
            return root.photo
        return None

    def resolve_avatar(root, info, **kwargs):
        if info.context.user.has_perm("core.view_avatar_rule", root):
            return root.avatar
        return None

    def resolve_avatar_url(root, info, **kwargs):
        if info.context.user.has_perm("core.view_avatar_rule", root) and root.avatar:
            return root.avatar.url
        return root.identicon_url

    def resolve_avatar_content_url(root, info, **kwargs):  # noqa
        # Returns the url for the main image for a person, either the avatar, photo or identicon,
        # based on permissions and preferences
        if get_site_preferences()["account__person_prefer_photo"]:
            if info.context.user.has_perm("core.view_photo_rule", root) and root.photo:
                return root.photo.url
            elif info.context.user.has_perm("core.view_avatar_rule", root) and root.avatar:
                return root.avatar.url

        else:
            if info.context.user.has_perm("core.view_avatar_rule", root) and root.avatar:
                return root.avatar.url
            elif info.context.user.has_perm("core.view_photo_rule", root) and root.photo:
                return root.photo.url

        return root.identicon_url

    def resolve_secondary_image_url(root, info, **kwargs):  # noqa
        # returns either the photo url or the avatar url,
        # depending on the one returned by avatar_content_url

        if get_site_preferences()["account__person_prefer_photo"]:
            if info.context.user.has_perm("core.view_avatar_rule", root) and root.avatar:
                return root.avatar.url
        elif info.context.user.has_perm("core.view_photo_rule", root) and root.photo:
            return root.photo.url
        return None

    def resolve_is_dummy(root: Union[Person, DummyPerson], info, **kwargs):
        return root.is_dummy if hasattr(root, "is_dummy") else False

    def resolve_notifications(root: Person, info, **kwargs):
        if root.pk and has_person(info.context) and root == info.context.user.person:
            return root.notifications.filter(send_at__lte=timezone.now()).order_by(
                "read", "-created"
            )
        return []

    def resolve_can_change_person_preferences(root, info, **kwargs):  # noqa
        return info.context.user.has_perm("core.change_person_preferences_rule", root)

    def resolve_can_impersonate_person(root, info, **kwargs):  # noqa
        return root.user and info.context.user.has_perm("core.impersonate_rule", root)

    def resolve_can_invite_person(root, info, **kwargs):  # noqa
        return (not root.user) and info.context.user.has_perm("core.invite_rule", root)

    @staticmethod
    def resolve_can_edit(root, info, **kwargs):
        if hasattr(root, "can_edit"):
            return root.can_edit
        return info.context.user.has_perm("core.edit_person_rule", root)

    @staticmethod
    def resolve_can_delete(root, info, **kwargs):
        if hasattr(root, "can_delete"):
            return root.can_delete
        return info.context.user.has_perm("core.delete_person_rule", root)

    @staticmethod
    def resolve_can_change_password(root, info, **kwargs):
        return info.context.user.has_perm(
            "core.change_password_rule", root
        ) or info.context.user.has_perm("core.change_user_password_rule", root)

    @staticmethod
    def resolve_can_send_password_reset_request(root, info, **kwargs):
        return info.context.user.has_perm("core.reset_user_password_rule", root)


class AddressInputType(graphene.InputObjectType):
    street = graphene.String(required=False)
    housenumber = graphene.String(required=False)
    postal_code = graphene.String(required=False)
    place = graphene.String(required=False)
    country = graphene.String(required=False)


class PersonInputType(graphene.InputObjectType):
    id = graphene.ID(required=False)  # noqa

    first_name = graphene.String(required=True)
    last_name = graphene.String(required=True)
    additional_name = graphene.String(required=False)
    short_name = graphene.String(required=False)

    street = graphene.String(required=False)
    housenumber = graphene.String(required=False)
    postal_code = graphene.String(required=False)
    place = graphene.String(required=False)

    phone_number = graphene.String(required=False)
    mobile_number = graphene.String(required=False)

    email = graphene.String(required=False)

    date_of_birth = graphene.Date(required=False)
    place_of_birth = graphene.String(required=False)
    sex = graphene.String(required=False)

    address = graphene.Field(AddressInputType, required=False)
    street = graphene.String(required=False)
    housenumber = graphene.String(required=False)
    postal_code = graphene.String(required=False)
    place = graphene.String(required=False)
    country = graphene.String(required=False)

    photo = Upload()
    avatar = Upload()

    guardians = graphene.List(lambda: PersonInputType, required=False)

    # TODO: Primary Group

    description = graphene.String(required=False)


class PersonBatchDeleteMutation(BaseBatchDeleteMutation):
    class Meta:
        model = Person
        permissions = ("core.delete_person_rule",)


class PersonGuardianMutationMixin:
    """Add functionality to handle guardians input to mutations."""

    @classmethod
    def _handle_guardians(cls, root, info, input, obj):  # noqa: A002
        """Handle and save guardian input."""
        guardian_role = Role.objects.get(short_name="guardian")

        if "guardians" in input:
            current_guardian_relationships = PersonRelationship.objects.filter(
                of_person=obj,
                roles=guardian_role,
            )

            handled_guardians = set()

            for relationship in current_guardian_relationships:
                if relationship.person.pk in input["guardians"]:
                    handled_guardians.add(relationship.person.pk)
                else:
                    relationship.delete()

            for new_guardian in list(set(input["guardians"]) - handled_guardians):
                relationship = PersonRelationship.objects.create(
                    person=Person.objects.get(pk=new_guardian),
                    of_person=obj,
                )
                relationship.roles.add(guardian_role)


class PersonAddressMutationMixin:
    """Add functionality to handle address input to mutations."""

    ADDRESS_FIELDS = ["street", "housenumber", "postal_code", "place", "country"]

    @classmethod
    def _handle_address(cls, root, info, input, obj):  # noqa: A002
        """Handle and save address input."""
        address_type = AddressType.get_default()

        if any((key in input) for key in cls.ADDRESS_FIELDS):
            try:
                address = obj.addresses.get(address_types=address_type)
            except Address.DoesNotExist:
                address = obj.addresses.create()
                address.address_types.add(address_type)

            for key in cls.ADDRESS_FIELDS:
                if key not in input:
                    continue
                setattr(address, key, input[key] or "")

            address.full_clean()
            address.save()

            # Delete address if no address fields are set
            if all(not getattr(address, key) for key in cls.ADDRESS_FIELDS):
                address.delete()


class PersonBatchCreateMutation(
    PersonAddressMutationMixin, PersonGuardianMutationMixin, BaseBatchCreateMutation
):
    class Meta:
        model = Person
        permissions = ("core.create_person_rule",)
        only_fields = (
            "user",
            "first_name",
            "last_name",
            "additional_name",
            "short_name",
            "phone_number",
            "mobile_number",
            "email",
            "date_of_birth",
            "place_of_birth",
            "sex",
            "description",
            "photo",
            "avatar",
            "primary_group",
            "related_persons",
            "related_to_persons",
        )
        optional_fields = (
            "additional_name",
            "short_name",
            "phone_number",
            "mobile_number",
            "email",
            "date_of_birth",
            "place_of_birth",
            "sex",
            "description",
            "photo",
            "avatar",
            "primary_group",
            "related_persons",
            "related_to_persons",
        )
        custom_fields = {
            "street": graphene.String(required=False),
            "housenumber": graphene.String(required=False),
            "postal_code": graphene.String(required=False),
            "place": graphene.String(required=False),
            "country": graphene.String(required=False),
            "guardians": graphene.List(graphene.ID, required=False),
        }

    @classmethod
    def handle_sex(cls, value, name, info):
        if value is None:
            return ""
        return value

    @classmethod
    def after_create_obj(cls, root, info, input, obj, full_input):  # noqa: A002
        super().after_create_obj(root, info, input, obj, full_input)
        cls._handle_address(root, info, input, obj)
        cls._handle_guardians(root, info, input, obj)


class PersonBatchPatchMutation(
    PersonAddressMutationMixin, PersonGuardianMutationMixin, BaseBatchPatchMutation
):
    class Meta:
        model = Person
        permissions = ("core.edit_person_rule",)
        only_fields = (
            "id",
            "user",
            "first_name",
            "last_name",
            "additional_name",
            "short_name",
            "phone_number",
            "mobile_number",
            "email",
            "date_of_birth",
            "place_of_birth",
            "sex",
            "description",
            "photo",
            "avatar",
            "primary_group",
            "related_persons",
            "related_to_persons",
        )
        optional_fields = (
            "user",
            "first_name",
            "last_name",
            "additional_name",
            "short_name",
            "phone_number",
            "mobile_number",
            "email",
            "date_of_birth",
            "place_of_birth",
            "sex",
            "description",
            "photo",
            "avatar",
            "primary_group",
            "related_persons",
            "related_to_persons",
        )
        custom_fields = {
            "street": graphene.String(required=False),
            "housenumber": graphene.String(required=False),
            "postal_code": graphene.String(required=False),
            "place": graphene.String(required=False),
            "country": graphene.String(required=False),
            "guardians": graphene.List(graphene.ID, required=False),
        }

    @classmethod
    def handle_sex(cls, value, name, info):
        if value is None:
            return ""
        return value

    @classmethod
    def after_update_obj(cls, root, info, input, obj, full_input):  # noqa: A002
        super().after_update_obj(root, info, input, obj, full_input)
        if info.context.user.has_perm("core.change_person") or info.context.user.has_perm(
            "core.change_person", obj
        ):
            pass
        elif (
            has_person(info.context.user)
            and obj == info.context.user.person
            and get_site_preferences()["account__editable_fields_person"]
        ):
            for input_key in input:
                if (
                    input_key != "id"
                    and input_key not in get_site_preferences()["account__editable_fields_person"]
                ):
                    raise PermissionDenied(
                        "User not allowed to edit the given fields for own person."
                    )
            pass
        cls._handle_address(root, info, input, obj)
        cls._handle_guardians(root, info, input, obj)


class AccountRegistrationInputType(graphene.InputObjectType):
    from .user import UserInputType  # noqa

    person = graphene.Field(PersonInputType, required=True)
    user = graphene.Field(UserInputType, required=True)
    invitation_code = graphene.String(required=False)


class SendAccountRegistrationMutation(PersonAddressMutationMixin, graphene.Mutation):
    class Arguments:
        account_registration = AccountRegistrationInputType(required=True)

    ok = graphene.Boolean()

    @classmethod
    @transaction.atomic
    def mutate(cls, root, info, account_registration: AccountRegistrationInputType):
        # Initialize registering person to indicate that registration is in progress
        info.context._registering_person = None

        invitation = None

        if code := account_registration["invitation_code"]:
            formatted_code = "".join(code.lower().split("-"))
            try:
                invitation = PersonInvitation.objects.get(
                    key=formatted_code,
                )
            except PersonInvitation.DoesNotExist as exc:
                raise SuspiciousOperation from exc

        if not get_site_preferences()["auth__signup_enabled"] and not (
            get_site_preferences()["auth__invite_enabled"] and invitation
        ):
            raise PermissionDenied(_("Signup is not enabled."))

        # Create email
        email = None

        if invitation and invitation.email:
            email = invitation.email
        elif account_registration["user"] is not None:
            email = account_registration["user"]["email"]

        # Check username
        for validator in custom_username_validators:
            try:
                validator(account_registration["user"]["username"])
            except ValidationError as exc:
                raise ValidationError(_("This username is not allowed.")) from exc

        # Create user
        try:
            user = get_user_model().objects.create_user(
                username=account_registration["user"]["username"],
                email=email,
                password=account_registration["user"]["password"],
            )
        except IntegrityError as exc:
            raise ValidationError(_("A user with this username or e-mail already exists.")) from exc

        validate_password(account_registration["user"]["password"], user)

        # Create person if no invitation is given or if invitation isn't linked to a person
        if invitation and invitation.person:
            person = invitation.person
            person.email = email
            person.user = user
            person.save()
        else:
            try:
                person, created = Person.objects.get_or_create(
                    user=user,
                    defaults={
                        "email": email,
                        "first_name": account_registration["person"]["first_name"],
                        "last_name": account_registration["person"]["last_name"],
                    },
                )
            except IntegrityError as exc:
                raise ValidationError(
                    _("A person using the e-mail address %s already exists.") % email
                ) from exc

        # Store contact information in database
        person.email = email

        for field in Person._meta.get_fields():
            if (
                field.name in account_registration["person"]
                and account_registration["person"][field.name] is not None
                and account_registration["person"][field.name] != ""
            ):
                setattr(person, field.name, account_registration["person"][field.name])
        person.full_clean()
        person.save()

        # Store address information
        cls._handle_address(root, info, account_registration["person"], person)

        # Accept invitation, if exists
        if invitation:
            accept_invitation(invitation, info.context, info.context.user)
            inviter_role = get_site_preferences()["auth__inviter_role"]

            if inviter_role is not None and invitation.inviter is not None:
                relationship, __ = PersonRelationship.objects.get_or_create(
                    person=invitation.inviter.person,
                    of_person=person,
                )
                relationship.roles.add(inviter_role)

        person.notify_abount_account_registration()

        _act = Activity(
            title=_("You registered an account"),
            description=_(f"You registered an account with the username {user.username}"),
            app="Core",
            user=person,
        )

        # Store person in request to make it accessible for injected registration mutations
        info.context._registering_person = person

        return SendAccountRegistrationMutation(ok=True)



# Generated by Django 5.1.7 on 2025-04-03 11:42

import aleksis.core.mixins
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0081_migrate_announcements'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RemoveField(
            model_name='announcement',
            name='id',
        ),
        migrations.AlterField(
            model_name='announcement',
            name='calendarevent_ptr',
            field=models.OneToOneField(auto_created=True, null=False, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='core.calendarevent'),
        ),
        migrations.RemoveField(
            model_name='announcement',
            name='extended_data',
        ),
        migrations.RemoveField(
            model_name='announcement',
            name='managed_by_app_label',
        ),
        migrations.RemoveField(
            model_name='announcement',
            name='valid_from',
        ),
        migrations.RemoveField(
            model_name='announcement',
            name='valid_until',
        ),
    ]



from xml.etree import ElementTree

from django.core.cache import cache
from django.http import Http404
from django.urls import resolve

from ...mixins import DAVResource
from .base import DAVMultistatus, DAVResponse, ElementHandler, NotFoundObject, QueryBase


class DAVEtag(ElementHandler):
    _class_name = "{DAV:}getetag"

    def process_xml(self, base, response):
        super().process_xml(base, response)
        objects = [response.obj] if response.obj is not None else []
        etag = response.resource.getetag(base.request._request, objects)
        base.current_xml.text = f'"{etag}"'


class DAVDisplayname(ElementHandler):
    _class_name = "{DAV:}displayname"

    def process_xml(self, base, response):
        super().process_xml(base, response)
        base.current_xml.text = response.resource.get_dav_verbose_name()


class DAVResourcetype(ElementHandler):
    _class_name = "{DAV:}resourcetype"

    def process_xml(self, base, response):
        super().process_xml(base, response)
        resource_types = ["{DAV:}collection"]
        if not getattr(response.resource, "_is_registry", False):
            resource_types += response.resource.dav_resource_types

        if response.obj is None:
            for res_type_attr in resource_types:
                ElementTree.SubElement(base.current_xml, res_type_attr)


class DAVCurrentUserPrincipal(ElementHandler):
    _class_name = "{DAV:}current-user-principal"

    def process_xml(self, base, response):
        super().process_xml(base, response)
        href = ElementTree.SubElement(base.current_xml, "{DAV:}href")
        person_pk = base.request._request.user.person.pk
        href.text = f"/dav/contact/person/{person_pk}.vcf"  # FIXME: Implement principals


class DAVGetcontenttype(ElementHandler):
    _class_name = "{DAV:}getcontenttype"

    def process_xml(self, base, response):
        resource = response.resource
        if response.obj is not None:
            resource = response.obj
        content_type = resource.get_dav_content_type()

        if not content_type:
            raise Http404

        super().process_xml(base, response)
        base.current_xml.text = content_type


class DAVGetcontentlength(ElementHandler):
    _class_name = "{DAV:}getcontentlength"

    def process_xml(self, base, response):
        obj = response.obj
        if obj is None:
            raise Http404

        super().process_xml(base, response)

        key = f"{obj._class_name}_{obj.pk}_dav_contentlength"
        if (cached := cache.get(key, None)) is not None:
            contentlength = cached
        else:
            contentlength = len(response.resource.get_dav_file_content(
                base.request._request, [obj]
            ))
            cache.set(key, contentlength, 60 * 60 * 24)

        base.current_xml.text = str(contentlength)


class DAVSupportedReportSet(ElementHandler):
    _class_name = "{DAV:}supported-report-set"

    def process_xml(self, base, response):
        super().process_xml(base, response)

        supported_reports = [
            ("urn:ietf:params:xml:ns:caldav", "calendar-multiget"),
            ("urn:ietf:params:xml:ns:caldav", "calendar-query"),
        ]

        for r in supported_reports:
            supported_report = ElementTree.SubElement(base.current_xml, "{DAV:}supported-report")
            report = ElementTree.SubElement(supported_report, "{DAV:}report")
            ElementTree.SubElement(report, self._get_xml_sub(r))


class DAVCurrentUserPrivilegeSet(ElementHandler):
    _class_name = "{DAV:}current-user-privilege-set"

    def process_xml(self, base, response):
        super().process_xml(base, response)

        privileges = [
            ("DAV:", "read"),
        ]

        for p in privileges:
            privilege = ElementTree.SubElement(base.current_xml, "{DAV:}privilege")
            ElementTree.SubElement(privilege, self._get_xml_sub(p))


class DAVHref(ElementHandler):
    _class_name = "{DAV:}href"

    def post_handle(self):
        res = resolve(self.content)
        name = res.kwargs.get("name")
        pk = res.kwargs.get("id")

        resource = DAVResource.registered_objects_dict[name]
        try:
            obj = resource.get_objects(self.request._request).get(pk=pk)
        except resource.DoesNotExist:
            obj = NotFoundObject(self.content)

        self.request.objects = list(self.request.objects)
        self.request.objects.append(obj)


class DAVProp(ElementHandler):
    _class_name = "{DAV:}prop"


class DAVPropname(ElementHandler):
    _class_name = "{DAV:}propname"
    invisible = True

    def process_xml(self, base: DAVMultistatus, response: DAVResponse = None):
        base.current_xml = ElementTree.SubElement(base.current_xml, DAVProp._class_name)

        response.resource._add_dav_propnames(base.current_xml)


class DAVAllprop(ElementHandler):
    _class_name = "{DAV:}allprop"
    invisible = True

    def pre_handle(self):
        for name in self.request.resource.dav_live_props:
            self.request.startElementNS(name, None, {})
            self.request.endElementNS(name, None)

    def process(self, stage: str, base: DAVMultistatus, response: DAVResponse = None):
        xml_element = base.current_xml.find(DAVProp._class_name)
        if xml_element is not None:
            base.current_xml = xml_element

        super().process(stage, base, response)

    def process_xml(self, base: DAVMultistatus, response: DAVResponse = None):
        base.current_xml = ElementTree.SubElement(base.current_xml, DAVProp._class_name)


class Propfind(QueryBase, ElementHandler):
    _class_name = "{DAV:}propfind"
    invisible = True

    def post_handle(self):
        super().post_handle()

        _is_registry = getattr(self.request.resource, "_is_registry", False)

        if not _is_registry or self.request.depth is None:
            for resource in filter(
                lambda r: not getattr(r, "_is_registry", False), self.request.resources
            ):
                try:
                    objs = resource.get_objects(self.request._request)
                except NotImplementedError:
                    objs = []

                self.request.objects += objs



from graphene_django import DjangoObjectType

from ..models import Holiday
from .base import (
    BaseBatchCreateMutation,
    BaseBatchDeleteMutation,
    BaseBatchPatchMutation,
    DjangoFilterMixin,
    PermissionsTypeMixin,
)


class HolidayType(PermissionsTypeMixin, DjangoFilterMixin, DjangoObjectType):
    class Meta:
        model = Holiday
        fields = ("id", "holiday_name", "date_start", "date_end")
        filter_fields = {
            "id": ["exact", "lte", "gte"],
            "holiday_name": ["icontains"],
            "date_start": ["exact", "lte", "gte"],
            "date_end": ["exact", "lte", "gte"],
        }


class HolidayBatchCreateMutation(BaseBatchCreateMutation):
    class Meta:
        model = Holiday
        permissions = ("core.create_holiday_rule",)
        only_fields = ("holiday_name", "date_start", "date_end")


class HolidayBatchDeleteMutation(BaseBatchDeleteMutation):
    class Meta:
        model = Holiday
        permissions = ("core.delete_holiday_rule",)


class HolidayBatchPatchMutation(BaseBatchPatchMutation):
    class Meta:
        model = Holiday
        permissions = ("core.edit_holiday_rule",)
        only_fields = ("id", "holiday_name", "date_start", "date_end")



# Generated by Django 5.1.6 on 2025-02-17 13:05

import aleksis.core.mixins
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models
from recurrence import serialize


def _migrate_recurrences(apps, schema_editor):
    CalendarEvent = apps.get_model('core', 'CalendarEvent')
    for event in CalendarEvent.objects.using(schema_editor.connection.alias).filter(recurrences__isnull=False):
        if not event.recurrences.rrules:
            continue
        rule = event.recurrences.rrules[0]
        # We only store until in the rrule_until field
        rule.until = None
        event.rrule = serialize(rule)
        if event.rrule.startswith("RRULE:"):
            event.rrule = event.rrule[6:]
        event.save()

class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('core', '0072_birthdayevent_view'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='calendarevent',
            name='datetime_start_before_end',
        ),
        migrations.RemoveConstraint(
            model_name='calendarevent',
            name='date_start_before_end',
        ),
        migrations.AddConstraint(
            model_name='calendarevent',
            constraint=models.CheckConstraint(condition=models.Q(('datetime_end__gte', models.F('datetime_start')), ('datetime_start__isnull', True), ('datetime_end__isnull', True), _connector='OR'), name='datetime_start_before_end'),
        ),
        migrations.AddConstraint(
            model_name='calendarevent',
            constraint=models.CheckConstraint(condition=models.Q(('date_end__gte', models.F('date_start')), ('date_start__isnull', True), ('date_end__isnull', True), _connector='OR'), name='date_start_before_end'),
        ),
        migrations.RunPython(_migrate_recurrences)
    ]



from graphene_django import DjangoObjectType

from ..models import Address, AddressType
from .base import (
    DjangoFilterMixin,
    PermissionsTypeMixin,
)


class AddressTypeType(PermissionsTypeMixin, DjangoFilterMixin, DjangoObjectType):
    class Meta:
        model = AddressType
        fields = ["id", "name"]


class AddressType(PermissionsTypeMixin, DjangoFilterMixin, DjangoObjectType):
    class Meta:
        model = Address
        fields = ["id", "address_types", "street", "housenumber", "postal_code", "place", "country"]



from aleksis.core.models import CalendarWidget, ExternalLinkWidget, StaticContentWidget
from aleksis.core.schema.base import BaseBatchPatchMutation, BaseObjectType
from aleksis.core.schema.dashboard.dashboard import DashboardWidgetType


class ExternalLinkWidgetType(BaseObjectType):
    class Meta:
        model = ExternalLinkWidget
        interfaces = (DashboardWidgetType,)
        fields = ["url", "icon_url"]


class StaticContentWidgetType(BaseObjectType):
    class Meta:
        model = StaticContentWidget
        interfaces = (DashboardWidgetType,)
        fields = ["content"]


class CalendarWidgetType(BaseObjectType):
    class Meta:
        model = CalendarWidget
        interfaces = (DashboardWidgetType,)
        fields = ["selected_calendars"]


class CalendarWidgetBatchPatchMutation(BaseBatchPatchMutation):
    class Meta:
        model = CalendarWidget
        permissions = ("core.edit_dashboardwidget_rule",)
        fields = (
            "id",
            "status",
            "title",
            "selected_calendars",
        )
        return_field_name = "dashboardWidgets"


class StaticContentWidgetBatchPatchMutation(BaseBatchPatchMutation):
    class Meta:
        model = StaticContentWidget
        permissions = ("core.edit_dashboardwidget_rule",)
        fields = (
            "id",
            "status",
            "title",
            "content",
        )
        return_field_name = "dashboardWidgets"


class ExternalLinkWidgetBatchPatchMutation(BaseBatchPatchMutation):
    class Meta:
        model = ExternalLinkWidget
        permissions = ("core.edit_dashboardwidget_rule",)
        fields = (
            "id",
            "status",
            "title",
            "url",
            "icon_url",
        )
        return_field_name = "dashboardWidgets"


dashboard_types = [
    ExternalLinkWidgetType,
    StaticContentWidgetType,
    CalendarWidgetType,
]



from dataclasses import dataclass

from django.core.exceptions import PermissionDenied

import graphene
import qrcode
from allauth.account.internal.flows.reauthentication import (
    did_recently_authenticate,
    raise_if_reauthentication_required,
)
from allauth.mfa.adapter import DefaultMFAAdapter, get_adapter
from allauth.mfa.internal.flows.add import validate_can_add_authenticator
from allauth.mfa.models import Authenticator
from allauth.mfa.recovery_codes.internal.flows import (
    can_generate_recovery_codes,
    generate_recovery_codes,
)
from allauth.mfa.totp.internal import auth as totp_auth
from allauth.mfa.totp.internal import flows as totp_flows
from allauth.mfa.webauthn.internal import (
    auth as webauthn_auth,
)
from allauth.mfa.webauthn.internal import (
    flows as webauthn_flows,
)
from graphene import ObjectType
from qrcode.image.svg import SvgPathFillImage


class MFATOTPType(ObjectType):
    id = graphene.ID(required=False)
    activated = graphene.Boolean()
    secret = graphene.String(required=False)
    totp_url = graphene.String(required=False)
    totp_qr_code = graphene.String(required=False)

    def resolve_activated(root, info, **kwargs):
        return isinstance(root, Authenticator)

    def resolve_totp_qr_code(root, info, **kwargs):
        if not getattr(root, "totp_url", None):
            return None
        return qrcode.make(root.totp_url, image_factory=SvgPathFillImage).to_string(
            encoding="unicode"
        )


class MFARecoveryCodesType(ObjectType):
    unused_codes = graphene.List(graphene.String)
    unused_code_count = graphene.Int()
    total_code_count = graphene.Int()

    def resolve_total_code_count(root, info, **kwargs):
        return len(root.generate_codes())

    def resolve_unused_code_count(root, info, **kwargs):
        return len(root.get_unused_codes())

    def resolve_unused_codes(root, info, **kwargs):
        raise_if_reauthentication_required(info.context)
        return root.get_unused_codes()


class MFAWebauthnAuthenticatorType(ObjectType):
    id = graphene.ID()
    created_at = graphene.DateTime(required=False)
    last_used_at = graphene.DateTime()
    name = graphene.String()

    @staticmethod
    def resolve_id(root, info, **kwargs):
        return root.instance.id

    @staticmethod
    def resolve_created_at(root, info, **kwargs):
        return root.instance.created_at

    @staticmethod
    def resolve_last_used_at(root, info, **kwargs):
        return root.instance.last_used_at if root.instance.last_used_at else None


class MFAWebauthnType(ObjectType):
    credential_creation_options = graphene.JSONString()
    authenticators = graphene.List(MFAWebauthnAuthenticatorType)

    @staticmethod
    def resolve_credential_creation_options(root, info, **kwargs):
        validate_can_add_authenticator(info.context.user)
        creation_options = webauthn_flows.begin_registration(info.context, info.context.user, False)
        return creation_options

    @staticmethod
    def resolve_authenticators(root, info, **kwargs):
        return [
            a.wrap()
            for a in Authenticator.objects.filter(
                user=info.context.user,
                type=Authenticator.Type.WEBAUTHN,
            )
        ]


class TwoFactorType(ObjectType):
    totp = graphene.Field(MFATOTPType, required=False)
    recovery_codes = graphene.Field(MFARecoveryCodesType, required=False)
    did_recently_authenticate = graphene.Boolean()
    webauthn = graphene.Field(MFAWebauthnType)

    def resolve_did_recently_authenticate(root, info, **kwargs):
        return did_recently_authenticate(info.context)

    def resolve_totp(root, info, **kwargs):
        authenticator = Authenticator.objects.filter(
            type=Authenticator.Type.TOTP, user=info.context.user
        ).first()
        if not authenticator:
            err = validate_can_add_authenticator(info.context.user)
            if err:
                raise ValueError()
            adapter: DefaultMFAAdapter = get_adapter()
            secret = totp_auth.get_totp_secret(regenerate=True)
            totp_url = adapter.build_totp_url(info.context.user, secret)
            return MFATOTPType(secret=secret, totp_url=totp_url)
        return authenticator

    def resolve_recovery_codes(root, info, **kwargs):
        authenticator = Authenticator.objects.filter(
            user=info.context.user,
            type=Authenticator.Type.RECOVERY_CODES,
        ).first()
        if not authenticator:
            return None
        return authenticator.wrap()

    @staticmethod
    def resolve_webauthn(root, info, **kwargs):
        return True


@dataclass
class ActivateTOTPInput:
    secret: str


class ActivateTOTPMutation(graphene.Mutation):
    class Arguments:
        code = graphene.String(required=True)

    authenticator = graphene.Field(MFATOTPType)

    @classmethod
    def mutate(cls, root, info, **data):
        validate_can_add_authenticator(info.context.user)

        secret = totp_auth.get_totp_secret(regenerate=False)
        code = data["code"]
        if not totp_auth.validate_totp_code(secret, code):
            raise get_adapter().validation_error("incorrect_code")

        authenticator = totp_flows.activate_totp(info.context, ActivateTOTPInput(secret=secret))[0]

        return ActivateTOTPMutation(authenticator=authenticator)


class DeactivateAuthenticatorMutation(graphene.Mutation):
    class Arguments:
        id = graphene.ID(required=True)

    success = graphene.Boolean()

    @classmethod
    def mutate(cls, root, info, **data):
        authenticator = Authenticator.objects.get(id=data["id"], user=info.context.user)

        if authenticator.type == Authenticator.Type.TOTP:
            totp_flows.deactivate_totp(info.context, authenticator)
        elif authenticator.type == Authenticator.Type.WEBAUTHN:
            webauthn_flows.remove_authenticator(info.context, authenticator)
        else:
            raise PermissionDenied()

        return DeactivateAuthenticatorMutation(success=True)


class GenerateRecoveryCodesMutation(graphene.Mutation):
    authenticator = graphene.Field(MFARecoveryCodesType)

    @classmethod
    def mutate(cls, root, info, **data):
        if not can_generate_recovery_codes(info.context.user):
            raise get_adapter().validation_error("cannot_generate_recovery_codes")
        authenticator = generate_recovery_codes(info.context)

        return GenerateRecoveryCodesMutation(authenticator=authenticator.wrap())


class AddSecurityKeyMutation(graphene.Mutation):
    class Arguments:
        name = graphene.String(required=True)
        credential = graphene.JSONString(required=True)

    authenticator = graphene.Field(MFAWebauthnAuthenticatorType)

    @classmethod
    def mutate(cls, root, info, **data):
        validate_can_add_authenticator(info.context.user)

        credential = data["credential"]
        webauthn_auth.parse_registration_response(credential)
        webauthn_auth.complete_registration(credential)

        authenticator = webauthn_flows.add_authenticator(
            info.context,
            name=data["name"],
            credential=data["credential"],
        )[0]

        return AddSecurityKeyMutation(authenticator=authenticator.wrap())
